mod content;
mod item_right_side;
mod jewel_right_side;
mod map_right_side;

use std::marker::PhantomData;
use std::sync::{Arc, Weak};

use anyhow::Result;
use engine::prelude::*;
use rpg_components::components::inventory::Inventory;
use rpg_components::items::ability_book::Ability;

use super::page_content::PageContent;
use super::traits::*;
use super::{content::Content, CharacterWindow, Page};
use item_right_side::ItemRightSide;
use jewel_right_side::JewelRightSide;
use map_right_side::MapRightSide;

pub struct InventoryPage<A: Ability + 'static> {
    close: Weak<Button>,

    engine: Arc<Engine>,
    hero: Entity,

    grid: Arc<Grid>,

    tooltip: Arc<Grid>,
    content: Arc<Grid>,

    modes: [Box<dyn PageContentWrapper>; 3],

    current_mode: usize,

    ability_marker: PhantomData<A>,
}

impl<A: Ability + 'static> InventoryPage<A> {
    pub fn new(
        engine: &Arc<Engine>,
        hero: Entity,
        reference: Weak<CharacterWindow>,
        close: &Arc<Button>,
    ) -> Result<Self> {
        let grid = Grid::new(engine.gui_handler().clone(), 2, 1, false)?;

        let left_base = GuiSnippet::from_str(
            engine.gui_handler(),
            include_str!("../../resources/inventory/left_side.xml"),
        )?;

        grid.attach(left_base.clone(), 0, 0, 1, 1)?;
        Self::setup_content_switch(&left_base, reference.clone())?;

        let tooltip = left_base.element("tooltip")?;
        let content = left_base.element("content")?;

        // items
        let item_mode = PageContent::<A, _>::new(
            Content::new::<_, Self>(engine, reference.clone(), {
                let engine = engine.clone();
                let hero = hero.clone();

                move || {
                    let mut data = Vec::new();

                    engine.on_scene(|scene| {
                        let hero_object = scene.entity(hero)?;
                        let inventory = hero_object.get_component::<Inventory<A>>()?;

                        data = inventory.iter_items().cloned().collect();

                        Ok(())
                    })?;

                    Ok(data)
                }
            })?,
            {
                let ui = GuiSnippet::from_str(
                    engine.gui_handler(),
                    include_str!("../../resources/inventory/items/tooltip.xml"),
                )?;

                let equip: Arc<Label> = ui.element("equip")?;
                equip.set_info_icon(&engine.controller_icon(ControllerButton::A)?)?;

                let salvage: Arc<Label> = ui.element("salvage")?;
                salvage.set_info_icon(&engine.controller_icon(ControllerButton::X)?)?;

                let socket: Arc<Label> = ui.element("socket")?;
                socket.set_info_icon(&engine.controller_icon(ControllerButton::Y)?)?;

                let switch_mode: Arc<Label> = ui.element("switch_mode")?;
                switch_mode.set_info_icon(&engine.controller_icon(ControllerButton::LeftStick)?)?;

                ui
            },
            ItemRightSide::new::<A>(
                engine,
                include_str!("../../resources/inventory/items/right_side.xml"),
                &reference,
                hero,
            )?,
        );

        // jewels
        let jewel_mode = PageContent::<A, _>::new(
            Content::new::<_, Self>(engine, reference.clone(), {
                let engine = engine.clone();
                let hero = hero.clone();

                move || {
                    let mut data = Vec::new();

                    engine.on_scene(|scene| {
                        let hero_object = scene.entity(hero)?;
                        let inventory = hero_object.get_component::<Inventory<A>>()?;

                        data = inventory.iter_jewels().cloned().collect();

                        Ok(())
                    })?;

                    Ok(data)
                }
            })?,
            {
                let ui = GuiSnippet::from_str(
                    engine.gui_handler(),
                    include_str!("../../resources/inventory/jewels/tooltip.xml"),
                )?;

                let socket: Arc<Label> = ui.element("socket")?;
                socket.set_info_icon(&engine.controller_icon(ControllerButton::A)?)?;

                let combine: Arc<Label> = ui.element("combine")?;
                combine.set_info_icon(&engine.controller_icon(ControllerButton::Y)?)?;

                let switch_mode: Arc<Label> = ui.element("switch_mode")?;
                switch_mode.set_info_icon(&engine.controller_icon(ControllerButton::LeftStick)?)?;

                ui
            },
            JewelRightSide::new::<A>(
                engine,
                include_str!("../../resources/inventory/jewels/right_side.xml"),
                hero,
                &reference,
            )?,
        );

        // maps
        let map_mode = PageContent::<A, _>::new(
            Content::new::<_, Self>(engine, reference.clone(), {
                let engine = engine.clone();
                let hero: Entity = hero.clone();

                move || {
                    let mut data = Vec::new();

                    engine.on_scene(|scene| {
                        let hero_object = scene.entity(hero)?;
                        let inventory = hero_object.get_component::<Inventory<A>>()?;

                        data = inventory.iter_maps().cloned().collect();

                        Ok(())
                    })?;

                    Ok(data)
                }
            })?,
            {
                let ui = GuiSnippet::from_str(
                    engine.gui_handler(),
                    include_str!("../../resources/inventory/maps/tooltip.xml"),
                )?;

                let select: Arc<Label> = ui.element("select")?;
                select.set_info_icon(&engine.controller_icon(ControllerButton::A)?)?;

                let start: Arc<Label> = ui.element("start")?;
                start.set_info_icon(&engine.controller_icon(ControllerButton::X)?)?;

                let switch_mode: Arc<Label> = ui.element("switch_mode")?;
                switch_mode.set_info_icon(&engine.controller_icon(ControllerButton::LeftStick)?)?;

                ui
            },
            MapRightSide::new(
                engine,
                include_str!("../../resources/inventory/maps/right_side.xml"),
            )?,
        );

        Ok(Self {
            close: Arc::downgrade(close),

            engine: engine.clone(),
            hero,

            grid,

            tooltip,
            content,

            modes: [
                Box::new(item_mode),
                Box::new(jewel_mode),
                Box::new(map_mode),
            ],

            current_mode: 0,

            ability_marker: PhantomData,
        })
    }

    fn switch_to_jewels(&mut self) -> Result<()> {
        self.current_mode = 1;
        self.update_page(true)
    }

    fn update_page(&mut self, select: bool) -> Result<()> {
        match self.current_mode {
            0 => println!("update item view"),
            1 => println!("update jewel view"),
            2 => println!("update map view"),

            _ => unreachable!(),
        }

        {
            let mode = &mut self.modes[self.current_mode];

            self.tooltip.attach(mode.tooltip().clone(), 0, 0, 1, 1)?;
            self.content
                .attach(mode.content_mut().base().clone(), 0, 0, 1, 1)?;
            self.grid
                .attach(mode.right_side_mut().base().clone(), 1, 0, 1, 1)?;

            mode.content_mut().update(&self.engine, self.hero)?;
            mode.right_side_mut().refresh(&self.engine, self.hero)?;

            if select {
                mode.content_mut().select()?;

                if mode.content_mut().is_empty() {
                    if let Some(close) = self.close.upgrade() {
                        close.select()?;
                    }
                }
            }
        }

        self.connect_character_page()?;

        Ok(())
    }

    fn connect_character_page(&mut self) -> Result<()> {
        let mode = &mut self.modes[self.current_mode];

        match self.current_mode {
            0 => {
                let right_side = mode.right_side_mut().base();
                let gloves: Arc<Button> = right_side.element("gloves")?;
                let ring_0: Arc<Button> = right_side.element("ring_0")?;
                let ring_2: Arc<Button> = right_side.element("ring_2")?;
                let main_hand: Arc<Button> = right_side.element("main hand")?;

                let content_grid: Arc<Grid> = mode.content_mut().base().element("content")?;

                content_grid.connect(
                    ConnectDirection::East,
                    [
                        (0, gloves.selectable().unwrap()),
                        (1, ring_0.selectable().unwrap()),
                        (2, ring_2.selectable().unwrap()),
                        (3, main_hand.selectable().unwrap()),
                    ],
                )?;
            }
            1 => {
                let right_side = mode.right_side_mut().base();

                let reference: Arc<Button> = right_side.element("reference")?;
                let first: Arc<Button> = right_side.element("first")?;

                let content_grid: Arc<Grid> = mode.content_mut().base().element("content")?;

                content_grid.connect(
                    ConnectDirection::East,
                    [
                        (1, reference.selectable().unwrap()),
                        (0, reference.selectable().unwrap()),
                        (3, first.selectable().unwrap()),
                        (2, first.selectable().unwrap()),
                    ],
                )?;
            }
            2 => {
                // map tab is not implemented now
            }

            _ => unreachable!(),
        }

        Ok(())
    }

    fn setup_content_switch(
        left_base: &GuiSnippet,
        reference: Weak<CharacterWindow>,
    ) -> Result<()> {
        let switch = {
            let reference = reference.clone();

            move |index| {
                if let Some(menu) = reference.upgrade() {
                    let mut tabs = menu.tabs_mut();
                    let me = tabs.inventory::<A>();

                    if me.current_mode != index {
                        me.current_mode = index;
                        me.update_page(true)?;
                    }
                }

                Ok(())
            }
        };

        let switch_to_items = Box::new({
            let switch = switch.clone();

            move || switch(0)
        });

        let switch_to_jewels = Box::new({
            let switch = switch.clone();

            move || switch(1)
        });

        let switch_to_maps = Box::new({
            let switch = switch.clone();

            move || switch(2)
        });

        left_base.set_click_callbacks(vec![
            ("items", switch_to_items),
            ("jewels", switch_to_jewels),
            ("maps", switch_to_maps),
        ])
    }
}

impl<A: Ability + 'static> Page for InventoryPage<A> {
    fn enable(&mut self) -> Result<Arc<Grid>> {
        println!("enable InventoryPage");

        for mode in self.modes.iter_mut() {
            mode.content_mut().refresh()?;
        }

        self.update_page(false)?;

        Ok(self.grid.clone())
    }

    fn disable(&mut self) -> Result<()> {
        self.modes[self.current_mode]
            .right_side_mut()
            .disable(&self.engine, self.hero)
    }

    fn select(&self) -> Result<()> {
        let mode = &self.modes[self.current_mode];

        mode.content().select()?;

        if mode.content().is_empty() {
            if let Some(close) = self.close.upgrade() {
                close.select()?;
            }
        }

        Ok(())
    }

    fn next_tab(&mut self) -> Result<()> {
        {
            let mode = &mut self.modes[self.current_mode];

            mode.content_mut().next_tab(&self.engine, self.hero)?;
            mode.content_mut().select()?;
        }

        self.connect_character_page()
    }

    fn previous_tab(&mut self) -> Result<()> {
        {
            let mode = &mut self.modes[self.current_mode];

            mode.content_mut().previous_tab(&self.engine, self.hero)?;
            mode.content_mut().select()?;
        }

        self.connect_character_page()
    }

    fn event(&mut self, button: ControllerButton) -> Result<bool> {
        Ok(match button {
            ControllerButton::LeftStick => {
                self.modes[self.current_mode]
                    .right_side_mut()
                    .disable(&self.engine, self.hero)?;

                self.current_mode = (self.current_mode + 1) % self.modes.len();

                self.update_page(true)?;
                true
            }
            ControllerButton::RightStick => {
                // check if jewel page is open
                if self.current_mode == 1 {
                    if JewelRightSide::combine::<A>(&self.engine, self.hero)? {
                        JewelRightSide::clear(&self.engine);
                        self.update_page(true)?;
                    }

                    true
                } else {
                    false
                }
            }

            _ => false,
        })
    }
}
