use engine::prelude::*;

use anyhow::Result;
use cgmath::Vector3;

use cgmath::InnerSpace;

#[derive(Debug, Serialize, Deserialize)]
struct AbilityLocationSplit {
    start: Vector3<f32>,
    end: Option<Vector3<f32>>,
}

impl AbilityLocationSplit {
    fn new(p: Vector3<f32>) -> Self {
        Self {
            start: p,
            end: None,
        }
    }

    // sets end point and creates a new split at this position

    fn finalize(&mut self, p: Vector3<f32>) -> Self {
        self.end = Some(p);
        Self::new(p)
    }

    // calculates length, if end is not set it uses `p`

    fn length(&self, p: Vector3<f32>) -> f32 {
        let end = match self.end {
            Some(p) => p,
            None => p,
        };

        (end - self.start).magnitude()
    }
}

#[derive(Debug, Serialize, Deserialize)]
pub struct AbilityLocationInfo {
    splits: Vec<AbilityLocationSplit>,

    allowed_distance: f32,

    #[serde(skip)]
    audio: UnsafeComponentStore<Audio>,
}

impl AbilityLocationInfo {
    pub fn new(position: Vector3<f32>, distance: f32, audio: &mut Audio) -> Self {
        AbilityLocationInfo {
            splits: vec![AbilityLocationSplit::new(position)],

            allowed_distance: distance,

            audio: audio.into(),
        }
    }

    pub fn post_setup(&mut self, audio: &mut Audio) {
        self.audio = audio.into();
    }

    pub fn new_direction_at(&mut self, p: Vector3<f32>) {
        let new_split = self.splits.last_mut().unwrap().finalize(p);
        self.splits.push(new_split);
    }

    pub fn check_distance(&self, p: Vector3<f32>) -> bool {
        let mut total = 0.0;

        for split in self.splits.iter() {
            total += split.length(p);
        }

        total < self.allowed_distance
    }
}

impl EntityComponent for AbilityLocationInfo {
    fn enable(&mut self, _scene: &mut Scene) -> Result<()> {
        unsafe { self.audio.as_mut() }.play("idle_sound", true)?;

        Ok(())
    }

    fn name(&self) -> &str {
        Self::debug_name()
    }
}

impl ComponentDebug for AbilityLocationInfo {
    fn debug_name() -> &'static str {
        "AbilityLocationInfo"
    }
}
